/*
    libgpmi - General Package/Module Interface
    Copyright (C) 2005-2007 Patric 'TrueLight' Stout & Tibor 'Igor2' Palinkas
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/**
 * @file error.h
 * \brief Here are all errors stored, which GPMI can return.
 *
 * Remember that no functions have the type 'gpmi_errors' as return value, but they
 *  can still return an error-code like this. Per function is defined if this is true
 *  yes or no.
 *
 * The optional error stack can be enabled or disabled globally. Each thread
 * has a separate error stack.
 */

#ifndef GPMI_ERROR_H
#define GPMI_ERROR_H

#include <gpmi_config.h>

#ifndef NO_VARARG_MACRO
#	define gpmi_error(level, error_message...) gpmi_debug(level, error_message)
	//!< Show an error on a certain level (works via debug.h)
#else
#	define gpmi_error gpmi_debug
	//!< Show an error on a certain level (works via debug.h)
#endif

typedef struct gpmi_err_stack_t gpmi_err_stack_t;
struct gpmi_err_stack_t {
	int error;
	int from;
	char *details;
	gpmi_err_stack_t *next;
} ;

typedef enum gpmi_error_sources {
	gpmi_err_src_unknown           = 0,
	gpmi_err_src_kw_add            = 1,
	gpmi_err_src_kw_add_auto       = 2,
	gpmi_err_src_kw_find_multi     = 2,
	gpmi_err_src_kw_del            = 3,
	gpmi_err_src_libdep_load       = 4,
	gpmi_err_src_find_path         = 5,
	gpmi_err_src_load_library      = 6,
	gpmi_err_src_load_module       = 7,
	gpmi_err_src_pkg_load          = 8,
	gpmi_err_src_lib_load_lowlevel = 9,
} gpmi_error_source_t;

/**
 * Errors which you can get back from GPMI-functions.
 */
typedef enum gpmi_errors {
	gpmi_err_success            =  0,         //!< Everything is ok
	gpmi_err_load_library       = -34001,     //!< Failed to load the library
	gpmi_err_null_str           = -34002,     //!< Expected a non-null string, got a null-string
	gpmi_err_version            = -34003,     //!< Version mismatch
	gpmi_err_init_failed        = -34004,     //!< Initialization failed
	gpmi_err_too_long           = -34005,     //!< String is too long
	gpmi_err_already_loading    = -34006,     //!< Already loading the library
	gpmi_err_cyclic_dep         = -34007,     //!< Found cyclic dependency (load has been canceled)
	gpmi_err_kw_already_exists  = -34008,     //!< Key already exists, multiple records for the same key is not allowed
	gpmi_err_kw_not_autoinc     = -34009,     //!< Keyword table is not auto-increment
	gpmi_err_kw_not_int         = -34010,     //!< Keyword table is not integer type
	gpmi_err_kw_not_found	    = -34011,     //!< Keyword not found
	gpmi_err_file_not_found     = -34012,     //!< File not found
	gpmi_err_init_not_found     = -34013,     //!< Regcmd of a package is not found
	gpmi_err_entry_not_found    = -34014,     //!< Entry point not found (usually not an error, rather a warning)

	gpmi_invalid_int            = -48192,     //!< This is a special error, retreived from keywords, when the int was invalid
} gpmi_error_t;


void gpmi_err_stack_push(gpmi_error_t error, gpmi_error_source_t from, const char *details);	//!< Allocate and push an entry on the error stack
gpmi_err_stack_t *gpmi_err_stack_pop(void);	//!< pop an entry from the stack
void gpmi_err_stack_destroy(gpmi_err_stack_t *entry); //!< destroy and free and entry or the whole stack (if entry == NULL)

/* Enable/disable the error stack. They both return the previous state (0 = disabled, 1 = enabled) */
int gpmi_err_stack_enable(void);
int gpmi_err_stack_disable(void);

/* Helper routines: convert codes to strings for error messaged */
char *gpmi_str_err_stack_entry(gpmi_err_stack_t *entry); //!< convert an error stack entry to string
const char *gpmi_strerror(int gpmi_errno); //!< convert an error code to string
const char *gpmi_strerror_src(int srcno); //!< convert an error source to string
void gpmi_err_stack_process_str(void (*callback)(gpmi_err_stack_t *entry, char *string)); //!< call back a function with each entry of the error stack, converted to string

#endif /* GPMI_ERROR_H */
