/*
    libgpmi - General Package/Module Interface
    Copyright (C) 2005-2007 Patric 'TrueLight' Stout & Tibor 'Igor2' Palinkas
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/**
 * @file gpmi/keywords.c
 * \brief Handles keywords for GPMI.
 *
 * Mostly used for script-modules.
 */

#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>
#include "keywords.h"
#include "keyword_hash.h"
#include "hash.h"
#include "hash_cursor.h"
#include "global.h"
#include "error.h"
#include "threads.h"

/* This keyword type modifier is a valid for gpmi_keyword_create and creates
   a static keyword array. Intended for internal use only, this why it's a
   macro. */
#define gpmi_keyword_typemod_static		1024

/* Sometimes Igor2 is just lazy... don't worry about it ;) */
#define foreach_array(array) for(; array->name != NULL; array++)

/**
 * \brief Imports a static keyword array into a keyword hash.
 */
gpmi_keyword *gpmi_keyword_import(const char *table_name, gpmi_keyword_types value_type, const void *array)
{
	gpmi_keyword_static_int     *array_int    = (gpmi_keyword_static_int *)   array;
	gpmi_keyword_static_string  *array_string = (gpmi_keyword_static_string *)array;
	gpmi_keyword_static_ptr     *array_ptr    = (gpmi_keyword_static_ptr *)   array;
	gpmi_keyword *kw;

	/* Create the new keyword table */
	kw = gpmi_keyword_create(table_name, value_type + gpmi_keyword_typemod_static);

	assert(kw != NULL);

	/* run the import cycle depending on the value type */
	switch(value_type) {
		case gpmi_keyword_type_integer:
			/* We import direct data - disable auto incrementing*/
			kw->end = -1;

			foreach_array(array_int) {
				//We can do this because array_int is static!
				hash_store(kw->hash,     array_int->name, &(array_int->value));
				hash_store(kw->hash_rev, &(array_int->value), array_int->name);
			}
			break;
		case gpmi_keyword_type_string:
			foreach_array(array_string) {
				hash_store(kw->hash,     array_string->name, array_string->value);
				hash_store(kw->hash_rev, array_string->value, array_string->name);
			}
			break;
		case gpmi_keyword_type_pointer:
			foreach_array(array_ptr) {
				hash_store(kw->hash, array_ptr->name, array_ptr->value);
				hash_store(kw->hash, array_ptr->value, array_ptr->name);
			}
			break;
		default:
			assert(!"Invalid keyword type for import");
			exit(190);
	}

	/* Return a pointer to the keyword table */
	return kw;
}

/**
 * \brief Converts a static keyword table into a dynamic one
 */
void gpmi_keyword_convert_to_dynamic(gpmi_keyword *keyword_table)
{
	hash_cursor_t *hc;
	void *kw;
	int *old_data;
	void *new_data;


	/* Cycle through all the table */
	hc = hash_cursor_new(keyword_table->hash);

	for (kw = hash_cursor_next(hc); kw != NULL; kw = hash_cursor_next(hc)) {
		/* Read old data */
		old_data = hash_cursor_data(hc);

		/* Copy data to a freshly allocated storage */
		if (keyword_table->type == gpmi_keyword_type_integer) {
			new_data = malloc(sizeof(int));
			*(int *)new_data = *(int *)old_data;
		}
		else {
			new_data = strdup((char *)old_data);
		}

		/* Store new pointer */
		hash_cursor_data_write(hc, new_data);
	}

	/* Clean up the cursor */
	hash_cursor_destroy(hc);

	/* Mark keyword table as dynamic */
	keyword_table->flags |= gpmi_keyword_flag_malloced;

	/* Now a small hack: we have _static/_dynamic hash calculator pairs. We
	   garantee they calculate the hash value the same way, but _dynamic also
		 frees the record when it is destroyed. This allows us to replace the
		 hash calculator on the fly, without rehashing. */

	/* We do the tricky casting to keep table->hash void * in keywords.h so
	   we don't need to include hash.h there */
		if (keyword_table->type == gpmi_keyword_type_integer)
			((hash_t *)keyword_table->hash)->hash_calc = gpmi_hash_int_dynamic;
		else
			((hash_t *)keyword_table->hash)->hash_calc = gpmi_hash_str_dynamic;
}

/**
 * \brief Create a keyword-table.
 *
 * Negative value_type means static version of the type - this is intended
 * for internal use, for static keyword import.
 */
gpmi_keyword *gpmi_keyword_create(const char *table_name, gpmi_keyword_types value_type)
{
	gpmi_keyword *kw;
	void *calc, *calc_rev;

	assert(table_name != NULL);

	/* Allocate the keyword table */
	kw = malloc(sizeof(gpmi_keyword));

	assert(kw != NULL);

	/* Determine the right hasher function for the given value type */
	switch (value_type) {
		case gpmi_keyword_type_integer:
			calc      = gpmi_hash_int_dynamic;
			calc_rev  = gpmi_hash_int_rev;
			kw->end   = 0; // allow auto-increment
			kw->flags = gpmi_keyword_flag_malloced;
			break;
		case gpmi_keyword_type_string:
			calc      = gpmi_hash_str_dynamic;
			calc_rev  = hash_defcalc;
			kw->end   = -1; // disallow auto-increment
			kw->flags = gpmi_keyword_flag_malloced;
			break;
		case gpmi_keyword_type_pointer:
			calc      = gpmi_hash_ptr;
			calc_rev  = gpmi_hash_ptr_rev;
			kw->end   = -1; // disallow auto-increment
			kw->flags = gpmi_keyword_flag_malloced;
			break;

		case gpmi_keyword_type_integer + gpmi_keyword_typemod_static:
			calc      = gpmi_hash_int_static;
			calc_rev  = gpmi_hash_int_rev;
			kw->end   = 0; // allow auto-increment
			kw->flags = 0;
			break;
		case gpmi_keyword_type_string + gpmi_keyword_typemod_static:
			calc      = hash_defcalc;
			calc_rev  = hash_defcalc;
			kw->end   = -1; // disallow auto-increment
			kw->flags = 0;
			break;
		case gpmi_keyword_type_pointer + gpmi_keyword_typemod_static:
			calc      = gpmi_hash_ptr;
			calc_rev  = gpmi_hash_ptr_rev;
			kw->end   = -1; // disallow auto-increment
			kw->flags = 0;
			break;
		default:
			/* Specified an invalid type */
			assert(!"Invalid value type for create");
			exit(191);
	}

	/* Initial setup */
	kw->type     = value_type > gpmi_keyword_typemod_static ? value_type - gpmi_keyword_typemod_static : value_type;
	kw->name     = strdup(table_name);
	kw->hash     = hash_create(64, calc);
	kw->hash_rev = hash_create(64, calc_rev);

	assert(kw->name != NULL && kw->hash != NULL && kw->hash_rev != NULL);

	gpmi_shared_lock;
	/* If we already have a keyword table list, add the new table */
	if (gpmi_shared.gpmi_keywords != NULL)
		gpmi_keyword_add(gpmi_shared.gpmi_keywords, table_name, kw);
	gpmi_shared_unlock;

	/* Return the pointer of the new keyword table */
	return kw;
}

/**
 * \brief Add a name/value pair in a keyword table.
 */
int gpmi_keyword_add(gpmi_keyword *keyword_table, const char *name, ...)
{
	int *i;
	char *s;
	void *v;
	va_list ap;

	assert(name != NULL);

	/* We add a direct data - disable auto incrementing for sure */
	keyword_table->end = -1;

	/* Prepare to read variable arg */
	va_start(ap, name);

	/* Do not allow multiple records for a single key */
	if (hash_find(keyword_table->hash, name) != NULL) {
		gpmi_err_stack_push(gpmi_err_kw_already_exists, gpmi_err_src_kw_add, name);
		return gpmi_err_kw_already_exists;
	}

	/* The keyword table's type tells what type of argument we expect */
	switch(keyword_table->type) {
		case gpmi_keyword_type_integer:
			/* Check if we are dynamic already */
			if ((keyword_table->flags & gpmi_keyword_flag_malloced) != gpmi_keyword_flag_malloced) {
				/* First time a dynamic value is added in a static (imported) array.
				   To avoid mixing static and dynamic behaviour, we convert the
					 whole table into dynamic */
				gpmi_keyword_convert_to_dynamic(keyword_table);
			}

			/* Add an allocated (dynamic) integer in the dynamic keyword array */
			i = malloc(sizeof(int));
			*i = va_arg(ap, int);
			hash_store(keyword_table->hash,     name, i);
			hash_store(keyword_table->hash_rev, i, name);
			break;
		case gpmi_keyword_type_string:
			if ((keyword_table->flags & gpmi_keyword_flag_malloced) != gpmi_keyword_flag_malloced) {
				/* See comment above */
				gpmi_keyword_convert_to_dynamic(keyword_table);
			}

			s = va_arg(ap, char *);
			hash_store(keyword_table->hash,     name, s);
			hash_store(keyword_table->hash_rev, s, name);
			break;
		case gpmi_keyword_type_pointer:
			v = va_arg(ap, void *);
			hash_store(keyword_table->hash,     name, v);
			hash_store(keyword_table->hash_rev, v, name);
			break;
	}

	va_end(ap);

	return 0;
}

/**
 * \brief Add a name in an auto-increment keyword table.
 */
int gpmi_keyword_add_auto(gpmi_keyword *keyword_table, const char *name)
{
	int *i;

	assert(name != NULL);

	/* We can do it only on auto-increment lists */
	if (keyword_table->end < 0) {
		gpmi_err_stack_push(gpmi_err_kw_not_autoinc, gpmi_err_src_kw_add_auto, name);
		return gpmi_err_kw_not_autoinc;
	}

	/* Field end shouldn't allow us to auto-increment for non-integer lists */
	assert(keyword_table->type == gpmi_keyword_type_integer);

	/* Do not allow multiple records for a single key */
	if (hash_find(keyword_table->hash, name) != NULL) {
		gpmi_err_stack_push(gpmi_err_kw_already_exists, gpmi_err_src_kw_add_auto, name);
		return gpmi_err_kw_already_exists;
	}


	i = malloc(sizeof(int));
	*i = keyword_table->end;

	hash_store(keyword_table->hash,     name, i);
	hash_store(keyword_table->hash_rev, i, name);

	keyword_table->end++;
	return *i;
}

/**
 * \brief Find an item from a keyword-integer-table as integer.
 */
int gpmi_keyword_find_int(gpmi_keyword *keyword_table, const char *name)
{
	const int *i;

	/* Check if everything is okay */
	assert(name != NULL);
	assert(keyword_table->type == gpmi_keyword_type_integer);

	/* Find item */
	i = hash_find(keyword_table->hash, name);

	/* If not found, return error-int */
	if (i == NULL)
		return gpmi_invalid_int;

	return *i;
}

/**
 * \brief Find an item by name or numeric from a keyword-integer-table.
 */
int gpmi_keyword_find_int_numeric(gpmi_keyword *keyword_table, const char *name)
{
	char *s;
	int i;

	assert(name != NULL);
	assert(keyword_table->type == gpmi_keyword_type_integer);

	/* Try if it's an integer */
	i = strtol(name, &s, 0);
	if (*s == '\0') {
		/* Valid integer */
		return i;
	}

	/* Try if it's a keyword */
	return gpmi_keyword_find_int(keyword_table, name);
}

/**
 * \brief Find an item from a keyword-string-table.
 */
const char *gpmi_keyword_find_string(gpmi_keyword *keyword_table, const char *name)
{
	assert(name != NULL);
	assert(keyword_table->type == gpmi_keyword_type_string);

	return (const char *)hash_find(keyword_table->hash, name);
}

/**
 * \brief Find an item from a keyword-pointer-table.
 */
void *gpmi_keyword_find_ptr(gpmi_keyword *keyword_table, const char *name)
{
	assert(name != NULL);
	assert(keyword_table->type == gpmi_keyword_type_pointer);

	return (void *)hash_find(keyword_table->hash, name);
}

/**
 * \brief Find-reverse-function, to find the name of an item that is in a keyword-list.
 */
const char *gpmi_keyword_find_rev(gpmi_keyword *keyword_table, ...)
{
	int i;
	const char *res;
	va_list ap;

	/* Prepare to read variable arg */
	va_start(ap, keyword_table);

	/* The keyword table's type tells what type of argument we expect */
	switch(keyword_table->type) {
		case gpmi_keyword_type_integer:
			i = va_arg(ap, int);
			res = (const char *)hash_find(keyword_table->hash_rev, &i);
			break;
		case gpmi_keyword_type_string:
			res = (const char *)hash_find(keyword_table->hash_rev, va_arg(ap, char *));
			break;
		case gpmi_keyword_type_pointer:
			res = (const char *)hash_find(keyword_table->hash_rev, va_arg(ap, void *));
			break;
		default:
			/* Corrupted type field */
			assert(!"Invalid keyword type for find");
			exit(192);
	}

	va_end(ap);

	return res;
}

/**
 * \brief Find multiple items from a keyword-integer-list, where you can use + and stuff to indicate which items you want.
 */
int gpmi_keyword_find_multi(gpmi_keyword *array, const char *key, int *result)
{
	const char *current, *word_start;
	char tmp[128];
	int len, keyword;

	assert(key != NULL && array != NULL);

	*result = 0;

	/* It works only for integer arrays */
	if (array->type != gpmi_keyword_type_integer) {
		gpmi_err_stack_push(gpmi_err_kw_not_int, gpmi_err_src_kw_find_multi, key);
		return gpmi_err_kw_not_int;
	}

	/* Start parsing, first word starts at the beginning of key */
	for (word_start = key, current = key; ; current++) {
		switch (*current) {
			case '+':
			case '|':
			case ',':
			case '/':
			case '\0':
				/* Found a separator, copy out the keyword to tmp*/
				len = current - word_start;
				if (len > 126)
					len = 126;
				strncpy(tmp, word_start, len);
				tmp[len] = '\0';

				/* Resolve the keyword from tmp*/
				keyword = gpmi_keyword_find_int(array, tmp);
				if (keyword < 0) {
					gpmi_err_stack_push(gpmi_err_kw_not_found, gpmi_err_src_kw_find_multi, tmp);
					return gpmi_err_kw_not_found;
				}
				*result |= keyword;

				/* Exit if key ends */
				if ( *current == '\0')
					return 0;

				/* Set new word start assuming one character long separator */
				word_start = current + 1;
				break;
		}
	}

	/* We should never get here, but -Wall will warning without this return */
	return gpmi_invalid_int;
}

/**
 * \brief Reverse lookup for multiple items.
 *
 * @return Return format is in ITEM1+ITEM2+..
 */
const char *gpmi_keyword_find_rev_multi(gpmi_keyword *keyword_table, int bitfield)
{
	int n, allocated, used, len;
	char *res, *tmp;

	assert(keyword_table != NULL);

	/* This works only for integer types */
	if (keyword_table->type != gpmi_keyword_type_integer)
		return NULL;

	/* If the string is empty, return NULL */
	if (bitfield == 0)
		return NULL;

	/* Allocate 32 bytes, and start checking for bits */
	allocated = 32;
	res       = malloc(allocated);
	used      = 0;
	res[used] = '\0';
	n         = 1;

	assert(res != NULL);

	/* As long as bitfield aint empty, continue the search */
	while (bitfield != 0) {
		/* If the first bit is on, add the name to the 'res' */
		if ((bitfield & 1) == 1) {
			/* Find the name ('n' is the real number) */
			tmp = (char *)hash_find(keyword_table->hash_rev, &n);

			len = strlen(tmp);
			/* Check if 'res' is big enough, if not, add */
			if (allocated < len + used + 2) {
				allocated += len + 32;
				res = realloc(res, allocated);

				assert(res != NULL);
			}

			/* If this isn't the first keyword, add '+' */
			if (used != 0) {
				strcat((res + used), "+");
				used++;
			}

			/* Add the name */
			strcat((res + used), tmp);
			used += len;
		}

		/* Go to the next bit */
		bitfield >>= 1;
		n *= 2;
	}

	return res;
}

/**
 * \brief Delete an item from a keyword-table.
 *
 * Warning: before you delete a string or a pointer, you need to free it, if
 *  it was allocated.

 * For integers, if you add dynamic content to imported tables, deleting the
 * dynamic content will not cause frees, you have to free them manual.
 *
 */
int gpmi_keyword_del(gpmi_keyword *keyword_table, const char *name)
{
	void *v;

	assert(name != NULL);

	/* Do not delete a non-existing key */
	v = (void *)hash_find(keyword_table->hash, name);
	if (v == NULL) {
		gpmi_err_stack_push(gpmi_err_kw_not_found, gpmi_err_src_kw_del, name);
		return gpmi_err_kw_not_found;
	}

	/* Delete the keys - Important: first delete from _rev since normal hash delete may free the record*/
	hash_del_rec(keyword_table->hash_rev, v, (char *)name);
	hash_del_rec(keyword_table->hash,     (char *)name, v);

	return 0;
}

/**
 * \brief Destroy a keyword-table, free all memory.
 */
void gpmi_keyword_destroy(gpmi_keyword *kw)
{
	assert(kw != NULL);
	assert(kw->name != NULL);

	gpmi_shared_lock;
	/* If we already have a keyword table list, delete our table */
	if (gpmi_shared.gpmi_keywords != NULL)
		gpmi_keyword_del(gpmi_shared.gpmi_keywords, kw->name);
	gpmi_shared_unlock;

	free(kw->name);
	hash_destroy(kw->hash_rev);
	hash_destroy(kw->hash);

	/* Free table description */
	free(kw);
}


/**
 * \brief Initialize the use of keywords.
 */
void gpmi_keyword_init()
{
	gpmi_shared_lock;

	/* Don't do anything if the lib is already initialized */
	if (gpmi_shared.gpmi_keywords != NULL) {
		gpmi_shared_unlock;
		return;
	}

	gpmi_shared.gpmi_keywords = gpmi_keyword_create("Keyword tables", gpmi_keyword_type_pointer);
	gpmi_shared_unlock;
}

/**
 * \brief Uninitialize the use of keywords.
 */
void gpmi_keyword_uninit()
{
	gpmi_shared_lock;

	/* Don't do anything if the lib is not initialized */
	if (gpmi_shared.gpmi_keywords == NULL) {
		gpmi_shared_unlock;
		return;
	}

	gpmi_keyword_destroy(gpmi_shared.gpmi_keywords);
	gpmi_shared.gpmi_keywords = NULL;
	gpmi_shared_unlock;
}
