/*
    libgpmi - General Package/Module Interface - Guile module
    Copyright (C) 2007 Mate Nagy
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* GUILE GPMI module */

#include <libguile.h>
#include <guile/gh.h>

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <stdio.h>

#include "gpmi/debug.h"

#include "guile.h"


/**
 * Returns the extension of this script-module.
 */
char *Guile_GetExtension()
{
	return strdup(".scm");
}

/**
 * Reloads the script on the fly.
 */
int Guile_Rehash(void *client_data)
{
	Guile_ClientData *MyData = client_data;
	char *name;
	int r;

	name = MyData->name;
	if (!name)
		return 0;

	r = Guile_LoadScript(client_data, name);
	free(name);
	return r;
}

/**
 * Loads a script.
 */
int Guile_LoadScript(void *client_data, const char *filename)
{
	Guile_ClientData *MyData = client_data;
	MyData->name = strdup(filename);

	scm_set_current_module(MyData->module);
	gh_eval_file(filename);

	return 0;
}

/**
 * Execute a dynamic command.
 **/
SCM Guile_DynamicPackageCommand(SCM func, SCM rest)
{
	int argc, froffset, i;
	char **args;
	char *s;
	char buf[32];

	char * (* proc) (int, char **);

	SCM_ASSERT(SCM_STRINGP(func), func, SCM_ARG1, "guile-gpmi-exec");

	char *procname;
	procname = gh_scm2newstr(func, NULL);

	proc = GPMI_Script_FindProc((char *) procname);

	/* Check if command not found - this must be an internal error: the
	   package had registered the command in the hash, but the routine
	   pointer is NULL.
	*/
	if (proc == NULL) {
		free(procname);
		gpmi_debug(0, "[Guile] Internal error: %s has a NULL proc in the command hash\n", procname);
		return SCM_BOOL_F;
	}

	argc = scm_num2int(scm_length(rest), 0, "guile-gpmi-exec");

	args = (char **) malloc(sizeof(char *) * (argc+1));
	assert(args != NULL);

	froffset = argc;

	argc = 1;

	while(SCM_FALSEP(scm_null_p(rest))) {
		SCM val = SCM_CAR(rest);

		if (SCM_NFALSEP(scm_string_p(val))) {
			args[argc++] = gh_scm2newstr(val, NULL);
		} else if(SCM_NFALSEP(scm_integer_p(val))) {
			long ival = scm_num2long(val, 0, NULL);
			sprintf(buf, "%ld", ival);
			args[argc++] = strdup(buf);
		} else {
			double dval = scm_num2double(val, argc + 1, "guile-gpmi-exec");
			sprintf(buf, "%f", dval);
			args[argc++] = strdup(buf);
		}

		rest = SCM_CDR(rest);
	}

	args[0] = (char *) procname;

	s = proc(argc, args);

	/* free allocated strings */
	for(i = 1; i < argc; i++)
		free(args[i]);

	free(args);

	free(procname);

	if (!s) return SCM_BOOL_F;

	return scm_makfrom0str(s);
}

/**
 * Register a command in our command hash
 */
void Guile_RegisterCommand(void *client_data, const char *name)
{
	Guile_ClientData *MyData = client_data;
	Guile_FuncNames *fn;

	fn = (Guile_FuncNames *) (malloc(sizeof(Guile_FuncNames)));
	assert(fn != NULL);
	fn->name = strdup(name);
	fn->next = MyData->names;
	MyData->names = fn;

	/* register macro for guile-gpmi-exec */

	scm_set_current_module(MyData->module);

	scm_primitive_eval(scm_cons(gh_symbol2scm("guile-gpmi-declare"),
		scm_cons(gh_symbol2scm(name), SCM_EOL)));
}

/**
 * Call a function of the script by name, passing params.
 */
void Guile_RunFunction(void *client_data, const char *function, int argc, const char **argv)
{
	Guile_ClientData *MyData = client_data;
	int i;

	scm_set_current_module(MyData->module);

	SCM func_symbol = scm_c_lookup(function);
	SCM func = scm_variable_ref(func_symbol);

	SCM list = SCM_EOL;
	SCM listhead = list;

	for(i = 0; i < argc; i++) {
		if (SCM_CONSP(listhead)) {
			SCM_SETCDR(list, scm_cons(scm_makfrom0str(argv[i]), SCM_EOL));
			list = SCM_CDR(list);
		} else {
			list = listhead = scm_cons(scm_makfrom0str(argv[i]), SCM_EOL);
		}
	}

	gh_apply(func, listhead);
}

/* XXX: not thread safe. */
int _guile_module_count = 0; /* number of guile modules loaded */

int Guile_Initialize(void **client_data)
{
	Guile_ClientData *MyData;
	char buf[32];

	if (_guile_module_count == 0)
		scm_init_guile();

	MyData = malloc(sizeof(Guile_ClientData));
	memset(MyData, 0, sizeof(Guile_ClientData));

	sprintf(buf, "gpmi -%d", _guile_module_count);
	MyData->module = scm_c_define_module(buf, NULL, NULL);
	scm_set_current_module(MyData->module);

	scm_c_define_gsubr("guile-gpmi-exec", 1, 0, 1, Guile_DynamicPackageCommand);
	gh_eval_str("(defmacro guile-gpmi-declare (fname) `(define (,fname . rest) (apply guile-gpmi-exec ,(symbol->string fname) rest)))");

	*client_data = MyData;

	_guile_module_count++;

	return 0;
}

void Guile_Destroy(void **client_data)
{
	Guile_ClientData *MyData = *client_data;
	Guile_FuncNames *n, *m;

	if (MyData->name)
		free(MyData->name);

	for(n = MyData->names; n; ) {
		m = n;
		n = n->next;
		free(m->name);
		free(m);
	}

	/* Finally free up the client-data */
	free(*client_data);

	_guile_module_count--;
}
