/*
    libgpmi - General Package/Module Interface - gpmi_scripts package
    Copyright (C) 2005-2007 Patric 'TrueLight' Stout & Tibor 'Igor2' Palinkas
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <assert.h>
#include <gpmi.h>
#include <unistd.h>
#include "gpmi_scripts.h"
#include "scripts.h"

/* For dirname */
#ifdef WIN32
#include "gpmi/os_dep.h"
#else
#include <libgen.h>
#endif


char *(*gpmi_scripts_asm_scriptname)(const GPMI_Script_Info *info, const char *file_name) = NULL;

char *gpmi_scripts_asm_scriptname_default(const GPMI_Script_Info *info, const char *name)
{
	char *extension, *script_name, *s, *path, *search, *original_path;

	original_path = (gpmi_private.path_override == NULL ? gpmi_shared.path_scripts : gpmi_private.path_override);
	search        = strdup(original_path);
	script_name   = malloc(strlen(search) + strlen(name) + 32);
	extension     = info->get_extension();

	for(s = path = search; *s != '\0'; s++) {
		if ((*s == GPMI_PATH_SEP_CHAR)) {
			*s = '\0';
			sprintf(script_name, "%s/%s/main%s", path, name, extension);
			if (access(script_name, R_OK) == 0)
				goto end;
			path = s + 1;
		}
	}
	sprintf(script_name, "%s/%s/main%s", path, name, extension);
end:;
	gpmi_private.path_next = original_path + (s - search) + 1;
	free(extension);
	free(search);
	return script_name;
}

gpmi_package **GPMI_Script_loaded_pkg = NULL;

/**
 * Hasher routine for command hash
 */
static unsigned int GPMI_Script_CommandHash(int cmd, char *str, char *str2, void **result_ptr)
{
	char *s;
	unsigned char c;
	unsigned int a, n;

	switch(cmd) {

		case h2_cmd_key:
			for(a=0,n=0,s=str; (*s!=0) && (*s!=' ') && (*s!='\t') && (*s!='\n') && (*s!='\r'); s++,n++) {
				c = toupper(*s);
				a+=(((c - 48) % 128) <<(n));
			}
			return a;
			break;

		case h2_cmd_cmpkey:
			if (strcmp(str,str2) == 0) return 0;
			return 1;

		case h2_cmd_cmprec:
			/* records are (void *) procedure addresses */
			if (str == str2) return 0;
			return 1;

		case h2_cmd_allockey:
			s=strdup(str);
			*result_ptr = s;
			return 0;

		case h2_cmd_freekey:
			free(str);
			return 0;
	}

	return -1;
}

/**
 * Small inline function to convert int via layout to string, while dup'ing it
 */
static inline char* int_to_string_dup(const char *layout, int x)
{
	char tmp[1024];

	gpmi_snprintf(tmp, sizeof(tmp), layout, x);
	return strdup(tmp);
}

/**
 * Small inline function to convert pointer via layout to string, while dup'ing it
 */
static inline char* ptr_to_string_dup(void *x)
{
	char tmp[1024];

	gpmi_snprintf(tmp, sizeof(tmp), PRINTF_PTR, STD_PRINTFP_CAST x);
	return strdup(tmp);
}

static void GPMI_Script_Dynamic_Event(gpmi_module *mod, int event_id, int argc, gpmi_globev_param_t *argv_in)
{
	GPMI_Script_ClientData *MyData = mod->mod_data;
	GPMI_Script_EventData evdata;
	gpmi_globev_param_t *arg;
	char **argv;
	int i;

	/* XXX: optimize this: save this buffer per module */
	argv = malloc(sizeof(char *) * argc);

	/* Set the first item */
	argv[0] = int_to_string_dup("%d", event_id);

	/* Read all params */
	for (i = 1, arg = &(argv_in[0]); i < argc; i++, arg++) {
		switch (arg->type) {
			case gpmi_globev_type_string:
				if (arg->data.string != NULL)
					argv[i] = strdup(arg->data.string);
				else
					argv[i] = strdup("");
				break;
			case gpmi_globev_type_int:
				argv[i] = int_to_string_dup("%d", arg->data.integer);
				break;
			case gpmi_globev_type_uint:
				argv[i] = int_to_string_dup("%u", arg->data.uinteger);
				break;
			case gpmi_globev_type_uchar:
				argv[i] = int_to_string_dup("%u", arg->data.uchar);
				break;
			case gpmi_globev_type_keyword:
				argv[i] = int_to_string_dup("[%d]", arg->data.integer);
				break;
			case gpmi_globev_type_ptr:
				argv[i] = ptr_to_string_dup(arg->data.ptr);
				break;
			default:
			/* Default should generate error */
				argv[i] = strdup("[*unknown_type*]");
		}
	}

	/* Set the data to the event */
	evdata.client_data = MyData;
	evdata.argv        = argv;
	evdata.argc        = argc;
	module_events_call(MyData->events, event_id, GPMI_Script_Event_Callback, &evdata);

	/* Free the memory of the evdata */
	for (i = 0; i < argc; i++) {
		if (argv[i] != NULL)
			free(argv[i]);
	}

	free(argv);
}

void GPMI_Script_RegisterCommand(void *handle, char *name, void *proc_wrapper, void *proc_lowlevel, gpmi_package *pkg)
{
	GPMI_Script_ClientData *MyData = handle;
	char *function = strdup(name);

	/* Some script-modules NEED all function to be lower-case, else it fails.. make it so */
	if (MyData->info->function_lowercase != 0) {
		char *s;
		for (s = function; *s != '\0'; s++)
			*s = tolower(*s);
	}

	/* If the script-module needs the package handler, pass it */
	if (MyData->info->set_pkg != NULL)
		MyData->info->set_pkg(MyData->client_data, pkg);

	MyData->info->reg(MyData->client_data, function);

	hash_store(MyData->cmd_hash, function, proc_wrapper);

	free(function);
}

/* Generated by wrapgen */
void GPMI_PKG_GEN_gpmi_scripts_register_commands_(void (*regcmd)(void *handle, char *name, void *proc_wrapper, void *proc_lowlevel, gpmi_package *package), void *handle, void *package);

int GPMI_Script_Load(const GPMI_Script_Info *info, gpmi_module *module, const char *params)
{
	GPMI_Script_ClientData *MyData;
	char *tmp, tmp_path[PATH_MAX], *script_path;
	int ret;

	/* No script-name -> fail */
	if (params == NULL)
		return -1;

	/* Clean before using */
	module->mod_data = NULL;

	/* Allocate memory for our stuff */
	MyData = malloc(sizeof(GPMI_Script_ClientData));
	memset(MyData, 0, sizeof(GPMI_Script_ClientData));

	/* Try to initialize the script-engine */
	if (info->init(&MyData->client_data) != 0) {
		free(MyData);
		return -2;
	}

	/* Save module data before interpreting the script */
	module->mod_data = MyData;

	/* Safe the info-struct */
	MyData->info = info;

	/* Create the command hash table */
	MyData->cmd_hash = hash_create(128, GPMI_Script_CommandHash);

	/* Initialize our own commands */
	GPMI_PKG_GEN_gpmi_scripts_register_commands_(GPMI_Script_RegisterCommand, MyData, NULL);

	/* Set timers to NULL */
	MyData->timers    = NULL;
	MyData->timer_pkg = NULL;

	/* Assign all events */
	gpmi_event_bind_global(module, GPMI_Script_Dynamic_Event);

	/* Initialize local event table */
	MyData->events = module_events_init();

	/* Load the script */
	{

		/* If no user defined file name assambler, use the default one */
		if (gpmi_scripts_asm_scriptname == NULL)
			tmp = gpmi_scripts_asm_scriptname_default(info, params);
		else
			tmp = gpmi_scripts_asm_scriptname(info, params);

		/* Remember the directory of the current script, some languages don't allow easy lookup for this */
		strcpy(tmp_path, tmp);
		script_path = dirname(tmp_path);
		MyData->current_script_dir = strdup(script_path);

		ret = info->load_script(MyData->client_data, tmp);
		free(tmp);
		return ret;
	}
}

void GPMI_Script_FreeModuleEventData(void *data)
{
	free(data);
}

int GPMI_Script_CompareModuleEventData(void *data1, void *data2)
{
	return strcmp((char *)data1, (char *)data2) == 0;
}

void GPMI_Script_Destroy(gpmi_module *module)
{
	GPMI_Script_ClientData *MyData = module->mod_data;

	if (MyData == NULL)
		return;

	if (MyData->info->destroy != NULL)
		MyData->info->destroy(&MyData->client_data);

	if (MyData->timers != NULL) {
		timer_uninit(MyData->timers);
		gpmi_pkg_unload(MyData->timer_pkg);
	}

	module_events_uninit(MyData->events, GPMI_Script_FreeModuleEventData);

	hash_destroy(MyData->cmd_hash);

	free(MyData->current_script_dir);

	/* Unload all packages we might have loaded */
	{
		gpmi_package *pkg;

		while ((pkg = MyData->package) != NULL) {
			MyData->package = pkg->next;
			gpmi_pkg_unload(pkg);
		}
	}

	/* Clear memory and make sure we don't come here again without an initialize */
	free(MyData);
	module->mod_data = NULL;
}

const void *GPMI_Script_FindProc(const char *function)
{
	GPMI_Script_ClientData *MyData;
	gpmi_module *curr;

	curr = gpmi_get_current_module();
	MyData = curr->mod_data;
	return hash_find(MyData->cmd_hash, (char *)function);
}

void *GPMI_Script_GetClientData()
{
	GPMI_Script_ClientData *MyData;
	gpmi_module *curr;

	curr = gpmi_get_current_module();
	MyData = curr->mod_data;

	return MyData->client_data;
}

void GPMI_Script_RunTick(gpmi_module *module, int frame_counter)
{
	GPMI_Script_ClientData *MyData = module->mod_data;

	MyData->last_frame = frame_counter;

	/* Run timers if we have timers */
	if (MyData->timers != NULL)
		timer_run(MyData->timers, frame_counter);
}

void GPMI_Script_RunData(gpmi_module *module, void *data)
{
	GPMI_Script_ClientData *MyData = module->mod_data;
	char tmp[100];
	char *argv[1];
	char *function;

	/* Pass the pointer to rundata */
	gpmi_snprintf(tmp, sizeof(tmp), PRINTF_PTR, STD_PRINTFP_CAST data);
	argv[0] = strdup(tmp);

	/* Some script-modules NEED all function to be lower-case, else it fails.. make it so */
	if (MyData->info->function_lowercase != 0) {
		function = strdup("rundata");
	} else {
		function = strdup("RunData");
	}
	MyData->info->run_function(MyData->client_data, function, 1, (const char **)argv);

	free(argv[0]);
	free(function);
}

void GPMI_Script_Event_Callback(void *handle, void *data)
{
	GPMI_Script_EventData *evdata = handle;
	GPMI_Script_ClientData *MyData = evdata->client_data;
	char *procname = strdup((char *)data);

	/* Some script-modules NEED all function to be lower-case, else it fails.. make it so */
	if (MyData->info->function_lowercase != 0) {
		char *s;
		for (s = procname; *s != '\0'; s++)
			*s = tolower(*s);
	}

	MyData->info->run_function(MyData->client_data, procname, evdata->argc, (const char **)evdata->argv);

	free(procname);
}

void GPMI_Script_Timer_Callback(timer_cmds cmd, void *data, timer *timer)
{
	GPMI_Script_ClientData *MyData;
	GPMI_Script_TimerData *td;
	char *argv[1];
	char *function;
	char *s;
	gpmi_module *curr;

	curr = gpmi_get_current_module();
	MyData = curr->mod_data;

	td = timer->data;

	switch(cmd) {
		case TIMER_UNINIT:
			/* Free the script string */
			free(td->script);
			if (td->user_data != NULL)
				free(td->user_data);
			free(td);
			timer->data = NULL;
			break;

		case TIMER_TICK:
			/* Assamble a call script */
			 function = strdup(td->script);

			/* Some script-modules NEED all function to be lower-case, else it fails.. make it so */
			if (MyData->info->function_lowercase != 0) {

				for (s = function; *s != '\0'; s++)
					*s = tolower(*s);
			}

			td->TTL  = timer->TTL;
			td->freq = timer->frequency;

			argv[0] = ptr_to_string_dup(td);
			MyData->info->run_function(MyData->client_data, function, 1, (const char **)argv);
			free(argv[0]);
			free(function);
			break;

		case TIMER_INIT:
			/* Do nothing */
			break;
	}
}

void gpmi_puts(const char *s)
{
	puts(s);
	fflush(stdout);
}

void gpmi_puts_nonl(const char *s)
{
	printf("%s", s);
	fflush(stdout);
}
